/*

  Copyright (C) 1999-2005 by Alexander Tenbusch

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License 
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/


/* 

a simple memory manager with special debugging support

*/

#include <stdlib.h>
#include <stddef.h>
#include <stdio.h>
#include <string.h>
#include <assert.h>
#include <malloc.h>

#include <libxml/parser.h>

#include "tools.h"
#include "mem.h"

#define GARB_INIT 0xA3          /* init fill pattern for memory */

#define GARB_FREE 0x17          /* fill pattern for released memory */

#ifdef MDEBUG

/* 

warning: "! There is no realloc (resize) function" 

warning: "In der Shell 'export MALLOC_CHECK_=1' setzen und dann
Programm laufen lassen"

*/

#endif

/* block_t -- info record for one memory block */
typedef struct
  {
    void *ptr;                  /* pointer to memory block */

    size_t nmemb;               /* number of members */

    size_t size;                /* size of one member in byte */

    flag_t valid;               /* flag for valid memory */

#ifdef MDEBUG
    /* debug info about the sourcecode line of allocation */
    char caller[100];
#endif
  }
block_t;


/* mem_t -- record for an array of block_t */
typedef struct
  {
    block_t *list;              /* pointer to an array of block_t */

    size_t nmemb;               /* number of members */

    flag_t valid;               /* flag for valid mem_t */
  }
mem_t;


static mem_t m = {NULL, 0, false};                      /* global mem_t */

/* pointer to THE empty string */
xmlChar *empty_str = "";

/* function prototypes */

void mem_exit(void);


void
mem_init(void)
{
  /*

    initialization of m

  */

  assert( m.list == NULL );     /* no double init */

  if( atexit(mem_exit) != 0)    /* register mem_exit() for exit() */
    exit(EXIT_FAILURE);

  m.valid = true;
}


flag_t
mem_valid(void *ptr)
{
  /*

    validation of a pointer

  */

  size_t i;

  assert(m.valid == true);      /* m ist nitialized yet */

  i = m.nmemb;

  while ( i-- > 0)
    if (m.list[i].ptr == ptr)
      return m.list[i].valid;

  return false;
}


void *
#ifdef MDEBUG
mem_iregister(char *file, int line, void *ptr, size_t nmemb, size_t size)
#else
mem_register(void *ptr, size_t nmemb, size_t size)
#endif
{
  /*

    register a foreign pointer in m

  */

  void *nptr;

  assert(m.valid == true);      /* m ist nitialized yet */
  assert(ptr != NULL);
  assert(nmemb > 0 && size > 0);

  /* reallocation of m.list (old contents unchanged,
     BUT not the same place [pointer]) */
  nptr = realloc(m.list, (m.nmemb + 1) * sizeof(block_t));
  if (nptr == NULL)
    {
      free(ptr);
      fprintf(stderr, "! not enough memory ...\n");
      exit(EXIT_FAILURE);
    }
  m.list = (block_t *) nptr;

  /* registration of new memory block */
  m.list[m.nmemb].ptr = ptr;
  m.list[m.nmemb].size = size;
  m.list[m.nmemb].nmemb = nmemb;

#ifdef MDEBUG
  sprintf(m.list[m.nmemb].caller, "%s:%i", file, line);
#endif

  m.list[m.nmemb].valid = true;
  m.nmemb++;

  m.valid = true;

  return ptr;
}


void
mem_unregister(void *ptr)
{
  /*

    unregister, but not release, a pointer in m

  */

  int i;

  assert(ptr != NULL);

  i = (int) m.nmemb;

  while ( i-- > 0 ) /* all registered pointers */
    if (m.list[i].ptr == ptr)
      {
        m.list[i].valid = false;
        return;
      }

  assert(false);
}


void *
#ifdef MDEBUG
mem_ialloc(char *file, int line, size_t nmemb, size_t size)
#else
mem_alloc(size_t nmemb, size_t size)
#endif
{
  /*

    allocation of memory, similar to calloc()

    mem_ialloc() is the debug-version of mem_alloc(), additional
    information about the calling sourcecode line

    return value: valid pointer

  */

  void *ptr;                    /* pointer to new allocated memory */
  void *nptr;                   /* pointer for reallocation of m.list */

                                /* m ist initialized already */
  assert(m.valid == true);
                                /* valid values for number of members
                                   and size */
  assert(nmemb > 0 && size > 0);

  ptr = malloc(nmemb * size);   /* conventional memory allocation */
  if (ptr == NULL)              /* and checking */
    {
      fprintf(stderr, "! not enough memory ...\n");
      exit(EXIT_FAILURE);
    }

  /* reallocation of m.list (old contents unchanged,
     BUT not the same place [pointer]) */
  nptr = realloc(m.list, (m.nmemb + 1) * sizeof(block_t));
  if (nptr == NULL)
    {
      free(ptr);
      fprintf(stderr, "! not enough memory ...\n");
      exit(EXIT_FAILURE);
    }
#if 0
  if (m.list != NULL) {         /* erforderlich?? */
    free(m.list);
  }
#endif
  m.list = (block_t *) nptr;            /* extended list */

  /* registration of new memory block */
  m.list[m.nmemb].ptr = ptr;
  m.list[m.nmemb].size = size;
  m.list[m.nmemb].nmemb = nmemb;

#ifdef MDEBUG
  /* notes the sourcecode line of calling */
  sprintf(m.list[m.nmemb].caller, "%s:%i", file, line);
#endif

  /*! \todo GARB_INIT auf MDEBUG beschrnken */
  /* initialize the new memory block with trash */
  memset(m.list[m.nmemb].ptr, GARB_INIT,
         m.list[m.nmemb].size * m.list[m.nmemb].nmemb);

  m.list[m.nmemb].valid = true; /* set the valid flag for memory block */
  m.nmemb++;                    /* increment the number of members */

  m.valid = true;

  return ptr;
}

/*! unregister and release a pointer in m

*/
void
#ifdef MDEBUG
mem_ifree(char *file, const unsigned int line, const void *ptr)
#else
mem_free(void *ptr)
#endif
{
  size_t i;

  assert(m.valid == true);      /* m is initialized already */
  assert(ptr != NULL);

  i = m.nmemb;

  while ( i-- > 0 )                    /* all registered pointers */
    if (m.list[i].ptr == ptr)
      {
        assert(m.list[i].valid == true);

#ifdef MDEBUG
        /* test the access to the whole block and shred */
        memset(m.list[i].ptr, GARB_FREE, m.list[i].size);
#endif
        free(m.list[i].ptr);    /* release the block */
        mem_unregister( m.list[i].ptr );
        return;
      }

  /* if failed */

#ifdef MDEBUG
  fprintf(stderr, "unregistered mem_free(0x%X) at %s:%u\n", ptr, file, line);

  mem_info();
#endif

  assert(false);
}

/*! allocates a two dimensional array

\param nxmemb number of elements in first dimension
\param nymemb number of elements in second dimension
\param size of element

\return pointer to array
*/
void **
mem_alloc_2d(const size_t nxmemb, const size_t nymemb, const size_t size_element) 
{

#if 0
  void **ptr;           /* pointer to new allocated memory */
  void *ptr_b;

  size_t i;

  assert(nxmemb > 0 && nymemb > 0 && size_element > 0);

  ptr_b = mem_alloc(nxmemb * nymemb, size_element); /* block */

  ptr = (void **)mem_alloc(nxmemb, sizeof(void *));/* array of line pointers */

  ptr[0] = ptr_b;
  for (i=1; i<nxmemb; i++) {
    ptr[i] = ptr[i-1] + nymemb * size_element;
/*     ptr[i] = ptr_b; */
/*     ptr_b += nymemb * size_element; */
  }
#endif

  size_t i;
  char **ptr;           /* pointer to new allocated memory */
  char *ptr_b;
  /* 'void *' makes problems with BCC 5.5:
     "Size of the type 'void' is unknown or zero"  
  */

  assert(nxmemb > 0 && nymemb > 0 && size_element > 0);

  ptr_b = (char *)mem_alloc(nxmemb * nymemb, size_element); /* block */

  ptr = (char **)mem_alloc(nxmemb, sizeof(char *));/* array of line pointers */

  for (i=0; i<nxmemb; i++)
    ptr[i] = ptr_b + i * nymemb * size_element;

  return (void **)ptr;
}
/* End of mem_alloc_2d() */


/*! unregister and release a pointer in m
 */
void
mem_free_2d(void **ptr)
{

  assert(mem_valid(ptr));

  assert(mem_valid(*ptr));

  mem_free(*ptr);               /* release the block */

  mem_free(ptr);                /* release the pointer array */
}


void
mem_exit(void)
{
  /* releases the rest of all blocks at exit()-call  */

  size_t i;

  assert(m.valid == true);      /* m is initialized already */

  mem_info();

  for (i = 0; i < m.nmemb; i++) {
    if (m.list[i].valid) {
#ifdef MDEBUG
      mem_free( m.list[i].ptr );
#else
      free(m.list[i].ptr);
#endif
    }
  }

  free(m.list);
}


#ifdef MDEBUG

void
mem_info(void)
{
  /*

    debug information about the allocated blocks

  */

  size_t s;
  size_t j;
  size_t i;

  assert(m.valid == true);      /* m is initialized already */

  fprintf(stderr, "\nmem_info():\n\n" );

  for (s = 0, i = 0, j = 0; i < m.nmemb; i++)
    {
      fprintf(stderr,
              "%4i %-20s\t 0x%09X %8u x %8u = ",
              i,
              m.list[i].caller,
              m.list[i].ptr,
              (unsigned int) m.list[i].nmemb,
              (unsigned int) m.list[i].size );

      if (m.list[i].valid)
        {
          fprintf(stderr, "%8u\n",
                  (unsigned int) m.list[i].nmemb * m.list[i].size );
          s += m.list[i].size * m.list[i].nmemb;
          j += m.list[i].nmemb;
        }
      else
        fprintf(stderr, "released\n" );
    }

  fprintf( stderr,
           "___________________________________________________________\n"
           "\n not released:%45u\n\n",
           (unsigned int) s );
}

#endif
