/*

  Copyright (C) 1999-2005 by Alexander Tenbusch

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License 
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/


#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include <assert.h>

#include <libxml/parser.h>
#include <libxml/tree.h>

#include "tools.h"
#include "mem.h"
#include "petrinet.h"

/****************************************************************************

   Definition of private constants
*/

const xmlChar *tag_pnet = "PETRINET";

const xmlChar *tag_subnet = "SUBNET";

const xmlChar *tag_state = "STELLE";

const xmlChar *tag_transition = "TRANSITION";

const xmlChar *tag_edge = "RELATION";

const xmlChar *tag_hitem = "LINK";

/* const xmlChar *default_color = "black"; */

/****************************************************************************

   Declaration of private functions (prototypes)
*/

static flag_t
petrinet_states_new(petrinet_t *p);

static flag_t
petrinet_transitions_new(petrinet_t *p);

static flag_t
petrinet_edges_new(petrinet_t *p);

static edge_t *
petrinet_edges_init(petrinet_t *p, edge_t *current_edge, xmlNode * a_node);

static knot_t *
petrinet_knots_init(petrinet_t *p, knot_t *current_knot, xmlNode *a_node, const xmlChar *element_name);

static flag_t
petrinet_subnets_new(petrinet_t *p);

static subnet_t *
petrinet_subnets_init(petrinet_t *p, subnet_t *current_subnet, xmlNode *a_node);

static flag_t 
member_transition(petrinet_t *p, knot_t *t) ;

static flag_t 
member_state(petrinet_t *p, knot_t *s) ;

static void
set_path_out_flag(petrinet_t *p, flag_t f);

/****************************************************************************

   Definition of all functions
*/


/****************************************************************************

   Definition of functions
*/

/*!

\param filename_pnet name of an XML file

\return the pointer of the created petrinet
*/
petrinet_t *
petrinet_new(char *filename, run_t run_mode)
{
  petrinet_t *result;

  xmlDoc  *cfg;
/*   xmlNode *cfg_root_element; */
  xmlNode *doc_prefix_element;

  char filename_xml[BUFFER_LENGTH];
/*   char *executable; */

  xmlNode *xml_element;
  result = (petrinet_t *) mem_alloc(1, sizeof(petrinet_t));

  result->filename = filename;
  result->mode_run = run_mode;

  /* ---------------------------------------------------------------------------

  parse configuration xml

   */
  cfg = xmlParseFile(result->filename);
  printf_exit( (cfg == NULL), 2, "could not parse file '%s'", result->filename);
  result->cfg_root_element = xmlDocGetRootElement(cfg);

  /*! \todo check empty values */
  result->executable = (char *) attribute_str(child_element(result->cfg_root_element,"CGIBIN"), "CALL");

  if (result->mode_run == command_line) {
    result->filename_absolut = (char *) attribute_str(child_element(result->cfg_root_element,"CGIBIN"), "CFGPATH");
  }
  else {
    /* path is inside result->filename */
    result->filename_absolut = empty_str;
  }

  /* find doc_prefix 
   */
  doc_prefix_element = child_element(result->cfg_root_element,"DOCPREFIX");
  printf_exit((!doc_prefix_element),1,"There is no DOCPREFIX"); 

  if (result->mode_run == cgi && attribute_str(doc_prefix_element, "HREF")) {
    sprintf( result->doc_prefix,
	     "%s%s",
	     (char *) attribute_str(doc_prefix_element, "HREF"),
	     (char *) attribute_str(doc_prefix_element, "NAME"));
    result->html_prefix = empty_str;
  }
  else {
    sprintf( result->doc_prefix,
	     "%s",
	     (char *) attribute_str(doc_prefix_element, "NAME"));
    result->html_prefix = (char *) attribute_str(doc_prefix_element, "HTML");
  }

  /* find the name of petrinet xml
   */
  xml_element = child_element(result->cfg_root_element,"FILE");
  if (mode_cgi(result)) {
    sprintf( filename_xml,
	     "%s%s",
	     (char *) attribute_str(xml_element, "PATH"),
	     (char *) attribute_str(xml_element, "NAME"));
  }
  else {
    sprintf( filename_xml,
	     "%s",
	     (char *) attribute_str(xml_element, "NAME"));
  }

  printf_exit( (strlen(filename_xml) < 5), 7, "! No valid xml filename given");

  printf_debug("Read Petrinet from '%s'", filename_xml);

  /* ---------------------------------------------------------------------------

  parse the petrinet xml file and get the DOM tree

  */
  result->xml = xmlParseFile(filename_xml);

  printf_exit( (result->xml == NULL), 2, "could not parse file '%s'", filename_xml);

  /* Get the root element node */
  result->root_element = xmlDocGetRootElement(result->xml);

  /* set id of petrinet */
  result->id = attribute_str(result->root_element, "ID");

  printf_exit((result->id == NULL), 1, "No id found for '%s'", tag_pnet);

  /* set version of petrinet */
  result->version = attribute_str(result->root_element, "VERSION");

  /* search for the name of subnet -> children with content NAME */
  result->name =
    child_content(child_element(result->root_element, "NAME"));
      
  /* search for the description of subnet -> children with
     content DESCRIPTION */
  result->description =
    child_content(child_element(result->root_element, "DESCRIPTION"));

  /* search for the copyright of subnet -> children with
     content COPYRIGHT */
  result->copyright =
    child_content(child_element(result->root_element, "COPYRIGHT"));

  /* search for the contact of subnet -> children with
     content CONTACT */
  result->contact =
    child_content(child_element(result->root_element, "CONTACT"));

  /* search for the remark of subnet -> children with
     content REMARK */
  result->remark =
    child_content(child_element(result->root_element, "REMARK"));

  /* 

  */
  petrinet_subnets_new(result);

  petrinet_states_new(result);

  petrinet_transitions_new(result);

  petrinet_edges_new(result);

  result->path = (knot_t **) mem_alloc(result->n_s + result->n_t + 1, sizeof(knot_t*));

  result->l_path = ERROR_INDEX;

  result->start = NULL;

  result->target = NULL;

  return result;
}
/* End of petrinet_new() */


/*!

\param p pointer to the petrinet

\return success flag
*/
static flag_t
petrinet_subnets_new(petrinet_t *p)
{
  subnet_t *tail;

  assert(p->root_element);

  /* subnets */
  p->n_sn = number_of(p->root_element, tag_subnet, 0) + 1;

  printf_debug("Number of '%s' = %i", tag_subnet, p->n_sn);

  p->sn = (subnet_t *) mem_alloc(p->n_sn+1, sizeof(subnet_t));

  /*  p->sn->id = "all"; */

  tail = petrinet_subnets_init(p, p->sn, p->root_element);

  tail->id = NULL;              /* mark the end of state array */

#if 0
  /* set the index number */
  for (i=0; p->sn[i].id; i++) {
    assert (i < p->n_sn);
    p->sn[i].i = i;
  }
#endif

  return true;
}
/* End of petrinet_subnets_new() */


/*!

\param p pointer to the petrinet

\return success flag
*/
static flag_t
petrinet_states_new(petrinet_t *p)
{
  knot_t *tail;
  int i;

  assert(p->root_element);

  /* states */
  p->n_s = number_of(p->root_element, tag_state, 0);

  printf_debug("Number of '%s' = %i", tag_state, p->n_s);

  p->s = (knot_t *) mem_alloc(p->n_s+1, sizeof(knot_t));

  tail = petrinet_knots_init(p, p->s, p->root_element, tag_state);

  tail->id = NULL;              /* mark the end of state array */

  /* set the index number */
  for (i=0; p->s[i].id; i++) {
    assert (i < p->n_s);
    p->s[i].i = i;
  }

  return true;
}
/* End of petrinet_states_new() */


/*! 

\param p pointer to the petrinet

\return success flag
*/
static flag_t
petrinet_transitions_new(petrinet_t *p)
{
  knot_t *tail;
  int i;

  assert(p->root_element);

  /* transitions */
  p->n_t = number_of(p->root_element, tag_transition, 0);

  printf_debug("Number of '%s' = %i", tag_transition, p->n_t);

  p->t = (knot_t *) mem_alloc(p->n_t+1, sizeof(knot_t));

  tail = petrinet_knots_init(p, p->t, p->root_element, tag_transition);

  tail->id = NULL;              /* mark the end of transition array */

  /* set the index number */
  for (i=0; p->t[i].id; i++) {
    assert (i < p->n_t);
    p->t[i].i = i;
  }

  return true;
}
/* End of petrinet_transitions_new() */


/*!

\param p pointer to the petrinet

\return success flag
*/
static flag_t
petrinet_edges_new(petrinet_t *p)
{
  index_t i;
  index_t j;
  edge_t *tail;

  assert(p->root_element);

  /* edges */
  p->n_e = number_of(p->root_element, tag_edge, 0);

  printf_debug("Number of '%s' = %i", tag_edge, p->n_e);

  p->e = (index_t **) mem_alloc_2d( p->n_s, p->n_t, sizeof(index_t));

  p->edge = (edge_t *) mem_alloc( p->n_e+1, sizeof(edge_t));

  /* set all possible edges to 'none' */
  for (j=0; j < p->n_s; j++) {
    for (i=0; i < p->n_t; i++) {
      p->e[j][i] = 0;
    }
  }

  tail = petrinet_edges_init(p, p->edge, p->root_element);
  tail->from = NULL;
  tail->l = NULL;

  /*! \todo calculation of edges not ok */
  printf_debug("%.2f valid Edges", (tail - p->edge) / sizeof(edge_t));

  return true;
}
/* End of petrinet_edges_new() */


/*! initialize all found edges

\param p pointer to the petrinet
\param current_edge
\param a_node current node for recursive call

\return 
*/
static edge_t *
petrinet_edges_init(petrinet_t *p, edge_t *current_edge, xmlNode * a_node)
{
  xmlNode *current_node;

  for (current_node = a_node; current_node; current_node = current_node->next) {

    if (current_node->type == XML_ELEMENT_NODE
        && xmlStrEqual(current_node->name, "RELATION")) {
      /* current_node is a relation node */

      xmlChar *valid_str = attribute_str(current_node, "VALID");

      /* defaults */
      current_edge->l = NULL;
      current_edge->valid = false;

      if (valid_str && xmlStrEqual(valid_str, "no")) {
        /* no valid relation */
      }
      else {
        index_t a;
        index_t b;

        xmlChar *a_str = attribute_str(current_node, "FROM");
        xmlChar *b_str = attribute_str(current_node, "TO");
        xmlChar *color_str = attribute_str(current_node, "COLOR");
        xmlChar *style_str = attribute_str(current_node, "STYLE");
        xmlChar *descr_str = child_content(child_element(current_node, "DESCRIPTION"));

        if (a_str && b_str) {
          a = state_index4id(p, a_str);
          if (a == ERROR_INDEX) {
            /* it was a wrong assumtion,
               OR there is no attribute 'FROM'
               OR there is no valid stelle with id 'a_str'
            */
            a = transition_index4id(p, a_str);
            if (a == ERROR_INDEX) {
              /* ignore invalid relations */
	      printf_debug("Edge invalid FROM '%s' -> '%s'", a_str, b_str);
            }
            else {
              b = state_index4id(p, b_str);
              if (b != ERROR_INDEX) {
                /* b is a valid stelle, a is a valid transition */
                p->e[b][a] |= 2;

		current_edge->valid = true;
		current_edge->from  = &(p->t[a]);
		current_edge->to    = &(p->s[b]);
		current_edge->color = color_str;
		current_edge->style = style_str;
		current_edge->description = descr_str;
		current_edge->l = htlist_new(current_node, tag_hitem);
/* 		printf_debug("Edge OK '%s' -> '%s'", current_edge->from->id, current_edge->to->id); */
              }
            }
          }
          else {
            /* right assumtion */
            b = transition_index4id(p, b_str);

            if (b != ERROR_INDEX) {
              /* a is a valid stelle, b is a valid transition */
              p->e[a][b] |= 1;

	      current_edge->valid = true;
	      current_edge->from  = &(p->s[a]);
	      current_edge->to    = &(p->t[b]);
	      current_edge->color = color_str;
	      current_edge->style = style_str;
	      current_edge->description = descr_str;
	      current_edge->l = htlist_new(current_node, tag_hitem);
/* 	      printf_debug("Edge OK '%s' -> '%s'", current_edge->from->id, current_edge->to->id); */
	    }
            else {
              /* ignore invalid relations */
/* 	      printf_debug("Edge invalid TO '%s' -> '%s'", a_str, b_str); */
            }
          }
	}
      }
      current_edge++;
    }
    current_edge = petrinet_edges_init(p,current_edge,current_node->children);
  }

  return current_edge;
}


/*!

\param p pointer to the petrinet

\return pointer to the created subnet
*/
static subnet_t *
petrinet_subnets_init(petrinet_t *p, subnet_t *current_subnet, xmlNode *a_node)
{
  xmlNode *current_node;
  xmlChar *valid_str;

  for (current_node = a_node; current_node; current_node = current_node->next) {

    if (current_node->type == XML_ELEMENT_NODE
	&& (xmlStrEqual(current_node->name, tag_pnet)
	    || xmlStrEqual(current_node->name, tag_subnet))) {
      /* current_node is a subnet */

      /* set id of subnet */
      current_subnet->id = attribute_str(current_node, "ID");

      /* set valid flag */        
      valid_str = attribute_str(current_node, "VALID");
      current_subnet->valid = (valid_str && xmlStrEqual(valid_str, "no")) ? false : true;

      /* search for the name of subnet -> children with content NAME */
      current_subnet->name =
	child_content(child_element(current_node, "NAME"));
      
      /* search for the description of subnet -> children with
	 content DESCRIPTION */
      current_subnet->description =
	child_content(child_element(current_node, "DESCRIPTION"));

      current_subnet++;
    }

    current_subnet = petrinet_subnets_init(p, current_subnet, current_node->children);
  }

  return current_subnet;
}
/* End of petrinet_subnets_init() */


/*!

\param p pointer to the petrinet

\return 
*/
static knot_t *
petrinet_knots_init(petrinet_t *p, knot_t *current_knot, xmlNode *a_node, const xmlChar *element_name)
{
  xmlNode *current_node;
  xmlChar *valid_str;

  for (current_node = a_node; current_node; current_node = current_node->next) {

    if (current_node->type == XML_ELEMENT_NODE
	&& xmlStrEqual(current_node->name, element_name)) {
      /* current_node is a node w/ element_name */

      if (xmlStrEqual(current_node->name, tag_state)) {
	current_knot->type = state;
      }
      else if (xmlStrEqual(current_node->name, tag_transition)) {
	current_knot->type = transition;
      }
      else {
	continue;
      }

      /* set id of knot */
      current_knot->id = attribute_str(current_node, "ID");

/*       printf_debug("'%s' '%s'", current_node->name, current_knot->id); */

      /* set valid flag */        
      valid_str = attribute_str(current_node, "VALID");
      current_knot->valid = (valid_str && xmlStrEqual(valid_str, "no")) ? false : true;

      /* set color of knot */
      current_knot->color = attribute_str(current_node, "COLOR");

      /* set rating of knot */
      current_knot->rating = attribute_str(current_node, "RATING");

      current_knot->out = true;

      /* subnet */
      if (current_node->parent->type == XML_ELEMENT_NODE
	  && xmlStrEqual(current_node->parent->name, tag_subnet)) {
	/* parent of current_node is a subnet node */
	xmlChar *subnet_id = attribute_str(current_node->parent, "ID");
	/* find the pointer to this subnet_t */
	subnet_t *s;
	for (s=p->sn; s->id; s++) {
	  if (xmlStrEqual(subnet_id, s->id)) {
	    /* same id */
	    current_knot->sn = s;
	    /* break; */
	  }
	}
      }
      else {
	current_knot->sn = p->sn;
      }

      /* inheritance of valid flag */
      if (! current_knot->sn->valid) {
	current_knot->valid = false;
      }

      /* search for the name of knot -> children with content NAME */
      current_knot->name =
	child_content(child_element(current_node, "NAME"));
      
      /* search for the description of state -> children with
	 content DESCRIPTION */
      current_knot->description =
	child_content(child_element(current_node, "DESCRIPTION"));

      current_knot->l = htlist_new(current_node->children, tag_hitem); 

      current_knot++;
    }

    current_knot =
      petrinet_knots_init(p, current_knot, current_node->children, element_name);
  }

  return current_knot;
}
/* End of petrinet_knots_init() */

/*! release the allocated mem of a petrinet_t
*/
void
petrinet_free(petrinet_t *p)
{
  index_t i;

  for (i=0; i < p->n_s; i++)
    if (p->s[i].l)
      mem_free(p->s[i].l);
  mem_free(p->s);

  for (i=0; i < p->n_t; i++)
    if (p->t[i].l)
      mem_free(p->t[i].l);
  mem_free(p->t); 

  mem_free(p->sn);

  for (i=0; i <= p->n_e; i++)	/* including empty end marker link */
    if (p->edge[i].l)
      mem_free(p->edge[i].l);
  mem_free(p->edge);

  mem_free_2d((void **)p->e);

  mem_free(p->path);

  mem_free(p);

  return;
}
/* End of petrinet_free() */


index_t
edge_index4id(petrinet_t *p, xmlChar *id_from, xmlChar *id_to)
{
  /* edge_index4id() - searches the index number of 
  */

  index_t i;

  if (p && id_from && id_to) { 
    for (i=0; i < p->n_e; i++) {
      if (p->edge[i].valid
	  && (xmlStrEqual(p->edge[i].from->id, id_from) && xmlStrEqual(p->edge[i].to->id, id_to))) {
	return i;
      }
    }
  } 

  return ERROR_INDEX;
}
/* End of edge_index4id() */


index_t
state_index4id(petrinet_t *p, xmlChar *id)
{
  /* state_index4id() - searches the index number of 'id' in 'p->s[]'
  */

  index_t i;

  if (p && id) { 
    for (i=0; i < p->n_s; i++) {
      if (p->s[i].valid && strcmp(p->s[i].id, id) == 0) {
	return i;
      }
    }
  } 

  return ERROR_INDEX;
}
/* End of state_index4id() */


index_t
transition_index4id(petrinet_t *p, xmlChar *id)
{
  /* transition_index4id() - searches the index number of 'id' in 'p->t[]'
  */

  index_t i;

  if (p && id) {
    for (i=0; i < p->n_t; i++) {
      if (p->t[i].valid && strcmp(p->t[i].id, id) == 0) {
	return i;
      }
    }
  }

  return ERROR_INDEX;
}
/* End of transition_index4id() */


knot_t *
state_ptr4id(petrinet_t *p, xmlChar *id) 
{
  if (p && id) {
    index_t i = state_index4id(p,id);
    if (i > -1 && i < p->n_s) {
      return &(p->s[i]);
    }
  }

  return NULL;
}


knot_t *
transition_ptr4id(petrinet_t *p, xmlChar *id) 
{
  if (p && id) {
    index_t i = transition_index4id(p,id);
    if (i > -1 && i < p->n_t) {
      return &(p->t[i]);
    }
  }

  return NULL;
}


edge_t *
edge_ptr4id(petrinet_t *p, xmlChar *id_from, xmlChar *id_to)
{
  if (p && id_from && id_to) {
    index_t i = edge_index4id(p,id_from,id_to);
    if (i > -1 && i < p->n_e) {
      return &(p->edge[i]);
    }
  }

  return NULL;
}


index_t
subnet_index4id(petrinet_t *p, xmlChar *id)
{
  /* subnet_index4id() - searches the index number of 'id' in 'p->sn[]'
  */

  index_t i;

  if (p && id) { 
    for (i=0; i < p->n_sn; i++) {
      if (p->sn[i].valid && strcmp(p->sn[i].id, id) == 0) {
	return i;
      }
    }
  } 

  return ERROR_INDEX;
}
/* End of subnet_index4id() */


flag_t
edgep(petrinet_t *p, knot_t *a, knot_t *b)
{
  /* edgep() - 

  */

  assert(p);
  assert(a);
  assert(b);

  if (statep(a)) {
    if (transitionp(b)) {
      return (p->e[a->i][b->i] & 1);
    }
  }
  else if (transitionp(a)) {
    if (statep(b)) {
      return (p->e[b->i][a->i] & 2);
    }
  }

  return false;
}
/* End of edgep() */


static flag_t
member_transition(petrinet_t *p, knot_t *t) 
{

  index_t i;

  assert( p );

  if (transitionp(t)) {
    for (i=0; i < p->l_path; i++) {
      if ( transitionp(p->path[i]) && p->path[i] == t ) {
        return true;
      }
    }
  }

  return false;
}


static flag_t
member_state(petrinet_t *p, knot_t *s) 
{

  index_t i;

  assert( p );

  if (statep(s)) {
    for (i=0; i < p->l_path; i++) {
      if ( statep(p->path[i]) && p->path[i] == s ) {
        return true;
      }
    }
  }

  return false;
}


/*! sets the 'p->start' or 'p->target'

a transition as start makes no sense !
*/
flag_t
set_start_target(petrinet_t *p, int end, xmlChar *start_id)
{
  index_t s;

  assert(p);

  s = state_index4id(p,start_id);
  if (s == ERROR_INDEX) {
    printf_debug("'State %s' not valid", start_id);
    return false;
  }

  if (end) {
    printf_debug("Set end to '%s' -> %i", start_id, s);
    p->target = &(p->s[s]);
  }
  else {
    printf_debug("Set start to '%s' -> %i", start_id, s);
    p->start = &(p->s[s]);    
  }

  p->path[0] = p->start;
  p->l_path  = 0;
  p->n_paths = 0;

  return true;
}
/* End of set_start_target() */


/*! search for an new path 

p->start and p->target must be states!! see set_start_target()
*/
flag_t
path_forward(petrinet_t *p)
{

  index_t i;
  knot_t *l_knot;

  /*! \bug bidrectional relation is also a path */

  /*   assert(statep(p->start) || transitionp(p->start)); */
  /*   assert(statep(p->target) || transitionp(p->target)); */
  assert(statep(p->start));
  assert(statep(p->target));
  assert(p->path[0] == p->start);

  l_knot = p->path[p->l_path]; /* ptr to last knot */

  if (l_knot == p->target && p->l_path > 0) {
    /* target is reached, path is yet complete,
       probably branch_backward() found a short way */
    p->n_paths++;   /* count this path and */
    return true;    /* leave the recursion */
  }
  else if (statep(l_knot)) {
    /* last element is a state -> search for a transition knot */

    index_t l_state = state_index(l_knot);

    for (i=0; i < p->n_t; i++) { /* all transitions */
      if (p->t[i].valid
	  && edge_st(p,l_state,i)
          && (! member_transition(p, &(p->t[i])))) {
        /* found a valid transition,
	   with an edge to l_state and
	   not yet included in path */
        p->l_path++;
        p->path[p->l_path] = &(p->t[i]); /* enclose the pointer into path */
        return path_forward(p);	/* search recursive next knot for path[] */
      }
    }
  }
  else if (transitionp(l_knot)) {
    /* last element is a transition -> search for a state knot */

    index_t l_trans = transition_index(l_knot);

    for (i=0; i < p->n_s; i++) { /* all states */
      knot_t *s_ptr = &(p->s[i]);
      if (s_ptr->valid && edge_ts(p,l_trans,i)) {
	/* found a valid state with an edge to l_trans */
	if (member_state(p, s_ptr)) {
#if 1
	  if (s_ptr == p->start && s_ptr == p->target) {
	    /* special case: its a circle! */
	    p->l_path++;
	    p->path[p->l_path] = s_ptr;
	    p->n_paths++;   /* count this path and */
	    return true;    /* leave the recursion */
	  }
	  /* else: state included yet */
#endif
	}
	else {
	  /* not yet included in path */
	  p->l_path++;
	  p->path[p->l_path] = s_ptr;
	  if (p->path[p->l_path] == p->target) {
	    /* target is reached, path is completed */
	    p->n_paths++;   /* count this path and */

#if 0
  for (i=0; i <= p->l_path; i++) { /* all  */
    fprintf(stderr,"%s ", p->path[i]->id);
  }
  fprintf(stderr,"\n");
#endif

	    return true;    /* leave the recusion */
	  }
	  return path_forward(p);	/* search recursive next knot for path[] */
	}
      }
    }
  }
  else {
    assert(false);
  }

  return false;			/* nothing more found */
}
/* End of path_forward() */


/*! search for the last possible branch 
 */
flag_t
branch_backward(petrinet_t *p)
{

  index_t i;
  knot_t *ol_knot;
  knot_t *l_knot;

  assert(statep(p->start));
  assert(statep(p->target));
  assert(p->path[0] == p->start);

  if (p->l_path < 1) {
    /* can't change the start, nothing more found */
    return false;
  }

  ol_knot = p->path[p->l_path]; /* ptr to old last knot */

  p->l_path--;          /* one step backward */

  l_knot = p->path[p->l_path]; /* ptr to last knot */

  if (statep(l_knot)) {
    /* last element is a state -> search for a transition knot */

    index_t l_state = state_index(l_knot);
    index_t l_trans = transition_index(ol_knot); /* index of removed step */

    for (i=l_trans+1; i < p->n_t; i++) { /* next transitions after removed one */
      if (p->t[i].valid
	  && edge_st(p,l_state,i)
	  && ! member_transition(p, &(p->t[i]))) {
        /* found a valid transition,
	   with an edge to l_state and
	   not yet included in path */
        p->l_path++;
        p->path[p->l_path] = &(p->t[i]);
        return true;		/* leave backward recursion */
      }
    }
    return branch_backward(p);
  }
  else {
    /* last element is a transition */

    index_t l_trans = transition_index(l_knot);
    index_t l_state = state_index(ol_knot);

    for (i=l_state+1; i < p->n_s; i++) {
      if (p->s[i].valid
	  && edge_ts(p,l_trans,i)
	  && ! member_state(p, &(p->s[i]))) {
        /* found a state */
        p->l_path++;
        p->path[p->l_path] = &(p->s[i]);
        return true;
      }
    }
    return branch_backward(p);
  }

  assert(false);
}
/* End of branch_backward() */



int
path_next(void)
{
  /*  path_next() - 

  */

;

  return true;
}
/* End of path_next() */


flag_t
path_exist(petrinet_t *p, xmlChar *start_id, xmlChar *target_id)
{
  /*  path_exist() - 

  */

  if (strcmp(start_id,target_id) == 0) {
    index_t i = state_index4id(p,start_id);
    p->s[i].out = true;
    return true;
  }
  else if (petrinet_set_start(p, start_id) && petrinet_set_target(p, target_id)) {

    do {
      if (path_forward(p)) {
	index_t i;
	for (i=0; i<p->l_path; i++) {
	  p->path[i]->out = true;
	}
      }
    } while (branch_backward(p));

    return true;
  }
  else {
    return false;
  }

  assert(true);
}
/* End of path_exist() */


flag_t
petrinet_start_out(petrinet_t *p, xmlChar *s_id)
{
  /*

  petrinet_start_out() - 

  */
  knot_t *c;

  for (c=p->s; c->id; c++) {
    c->out = false;
  }

  for (c=p->t; c->id; c++) {
    c->out = false;
  }

  for (c=p->s; c->id; c++) {	/* all knots w/ id */
    if (c->valid) {
#if 0
/*       if (path_exist(p,c->id,s_id) || path_exist(p,s_id,c->id)) { */
      if (path_exist(p,s_id,c->id)) {
	c->out = true;
      }
      else {
	c->out = false;
      }
#endif
/*       path_exist(p,s_id,c->id); */
      path_exist(p,c->id,s_id); 
    }
  }

  return true;
}
/* End of petrinet_start_out() */


void
set_path_out_flag(petrinet_t *p, flag_t f)
{
  /* 
     set all knots in p->path on f
  */
  index_t i;

  for (i=0; i <= p->l_path; i++) {
    p->path[i]->out = f;
  }
}
/* End of set_path_out_flag() */


flag_t
petrinet_pathnet_out_flag(petrinet_t *p, flag_t f)
{
  /* 
     set all knots in paths between state  on f

  */

  if (p->start && p->start->valid
      && p->target && p->target->valid) {

    petrinet_all_out_flag(p,!f);

    do {
      if (path_forward(p)) {
	set_path_out_flag(p,f);
      }
    } while (branch_backward(p));

    printf_debug("%i path between '%s' and '%s' found",
	    p->n_paths, p->start->id, p->target->id);
  }

  return (p->n_paths > 0);
}
/* End of petrinet_pathnet_out_flag() */


void
petrinet_all_out_flag(petrinet_t *p, flag_t f)
{
  /*

  petrinet_all_out_flag() - 

  */

  knot_t *k;

  assert(p);

  for (k=p->s; k->id; k++) {
    if (k->valid) {
      k->out = f;
    }
  }

  for (k=p->t; k->id; k++) {
    if (k->valid) {
      k->out = f;
    }
  }
}
/* End of petrinet_all_out_flag() */


void
petrinet_subnet_valid(petrinet_t *p, subnet_t *sn, flag_t f)
{
  /*

  petrinet_subnet_valid() - sets the pointed subnet and all member knots to f

  */

  knot_t *k;

  assert(p);
  assert(sn);

  sn->valid = f;

  for (k=p->s; k->id; k++) {
    if (k->sn && k->sn == sn) {
      k->valid = f;
    }
  }

  for (k=p->t; k->id; k++) {
    if (k->sn && k->sn == sn) {
      k->valid = f;
    }
  }


}
/* End of petrinet_subnet_valid() */


void
petrinet_neighbors_out_flag(petrinet_t *p, knot_t *k, flag_t f)
{

  if (k->valid) {

    k->out = f;

    if (k->type == state) {
      /* 'k' is a state knot */
      index_t t_i;
      for (t_i=0; t_i < p->n_t; t_i++) {
	if (p->t[t_i].valid) {
	  if (edge_st(p,k->i,t_i)) {
	    p->t[t_i].out = f;
	  }
	  if (edge_ts(p,t_i,k->i)) {
	    p->t[t_i].out = f;
	  }
	}
      }
    }
    else {
      /* 'k' is a transition knot */
      index_t s_i;
      for (s_i=0; s_i < p->n_s; s_i++) {
	if (p->s[s_i].valid) {
	  if (edge_st(p,s_i,k->i)) {
	    p->s[s_i].out = f;
	  }
	  if (edge_ts(p,k->i,s_i)) {
	    p->s[s_i].out = f;
	  }
	}
      } 
    }
  }
}
/* End of petrinet_neighbors_out_flag() */


