/*

  Copyright (C) 1999-2005 by Alexander Tenbusch

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License 
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

*/


#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h>
#include <ctype.h>
#include <string.h>

#include <assert.h>
#include <limits.h>

#include <libxml/parser.h>
#include <libxml/tree.h>

#include <cgi-util.h>

#include "tools.h"
#include "mem.h"
#include "petrinet.h"
#include "graph.h"
#include "web.h"
#include "pkg2.h"


const xmlChar *tag_requirement = "REQUIREMENT";

const xmlChar *tag_req = "REQ";

const xmlChar *tag_check = "CHECK";

static flag_t
requirements_new(pkg2_t *pkg2);

static requirement_t *
requirements_init(pkg2_t *p, requirement_t *current_requirement, xmlNode *a_node, const xmlChar *element_name);

static flag_t
requirement_refs_new(pkg2_t *pkg2);

static req_t *
reqs_init(pkg2_t *pkg2, req_t *current_req, xmlNode *a_node, const xmlChar *element_name);


pkg2_t *
pkg2_new(char *filename, run_t run_mode)
{
  pkg2_t *result;
  xmlNode *current_node;

  result = (pkg2_t *) mem_alloc(1, sizeof(pkg2_t));

  result->pn = petrinet_new(filename,run_mode);

  /* default mode settings */
  result->mode = 0;
/*   if (run_mode == cgi) result->mode |= FLAG_CGI; */

  current_node = child_element(result->pn->cfg_root_element,"FILE");

  for (current_node = result->pn->cfg_root_element->children;
       current_node; 
       current_node = current_node->next) {

    if (current_node->type == XML_ELEMENT_NODE
        && xmlStrEqual(current_node->name, "MODULE")) {
      /* current_node is a module node */

      xmlChar *name_str = attribute_str(current_node, "NAME");

      printf_debug("Use module '%s'", name_str);

      /*! \todo improve bit logic for pkg2->mode */
      if (xmlStrEqual(name_str, "develop")) {
	result->mode |= FLAG_DEVELOP;
      }
      else if (xmlStrEqual(name_str, "links")) {
	result->mode |= FLAG_LINKS;
      }
      else if (xmlStrEqual(name_str, "docs")) {
	result->mode |= FLAG_DOCS;
      }
      else if (xmlStrEqual(name_str, "requirements")) {
	result->mode |= FLAG_REQUIREMENTS;
      }
      else if (xmlStrEqual(name_str, "images")) {
	result->mode |= FLAG_IMAGES;
      }
      else {
	printf_debug("Not supported module '%s'", name_str);
      }

    }
#if 0
    else if (current_node->type == XML_ELEMENT_NODE
	     && xmlStrEqual(current_node->name, "CGIBIN")) {
      /* current_node is a module node */
      executable = (char *) attribute_str(current_node, "CALL");
    }
    else if (current_node->type == XML_ELEMENT_NODE
	     && xmlStrEqual(current_node->name, "FILE")) {
      /* current_node is a module node */

      if (mode_cgi(result)) {
	sprintf( filename_xml,
		 "%s%s",
		 (char *) attribute_str(current_node, "PATH"),
		 (char *) attribute_str(current_node, "NAME"));
      }
      else {
	sprintf( filename_xml,
		 "%s",
		 (char *) attribute_str(current_node, "NAME"));
      }
    }
    else if (current_node->type == XML_ELEMENT_NODE
	     && xmlStrEqual(current_node->name, "DOC_PREFIX")) {
      /* current_node is a module node */
      if (mode_cgi(result)) {
	sprintf( doc_prefix,
		 "%s%s",
		 (char *) attribute_str(current_node, "HREF"),
		 (char *) attribute_str(current_node, "NAME"));
      }
      else {
	sprintf( doc_prefix,
		 "%s",
		 (char *) attribute_str(current_node, "NAME"));
      }
    }
#endif
  }
 
#if 0
  if (mode_cgi(result)) {
    sprintf( filename_xml,
	     "%s%s",
	     (char *) attribute_str(current_node, "PATH"),
	     (char *) attribute_str(current_node, "NAME"));
  }
  else {
    sprintf( filename_xml,
	     "%s",
	     (char *) attribute_str(current_node, "NAME"));
  }
#endif

  requirements_new(result);

  requirement_refs_new(result);

  return result;
}
/* Ende von pkg2_new() */


/*!

\param p pointer to the pkg2

\return success flag
*/
static flag_t
requirements_new(pkg2_t *pkg2)
{
  requirement_t *tail;

  assert(pkg2);
  assert(pkg2->pn->xml);
  assert(pkg2->pn->root_element);

  /* requirements */
  pkg2->n_r = number_of(pkg2->pn->root_element, tag_requirement, 0);

  printf_debug("Number of '%s' = %i", tag_requirement, pkg2->n_r);

  pkg2->r = (requirement_t *) mem_alloc(pkg2->n_r+1, sizeof(requirement_t));

  tail = requirements_init(pkg2, pkg2->r, pkg2->pn->root_element, tag_requirement);

  tail->id = NULL;              /* mark the end of requirement array */

  return true;
}
/* End of pkg2_requirements_new() */


/*!

\param p pointer to the pkg2

\return 
*/
static requirement_t *
requirements_init(pkg2_t *pkg2, requirement_t *current_requirement, xmlNode *a_node, const xmlChar *element_name)
{
  xmlNode *current_node;
  xmlChar *valid_str;

  for (current_node = a_node; current_node; current_node = current_node->next) {

    if (current_node->type == XML_ELEMENT_NODE
	&& xmlStrEqual(current_node->name, element_name)) {
      /* current_node is a node w/ element_name */

      /* set id of requirement */
      current_requirement->id = attribute_str(current_node, "ID");

      /* set valid flag */        
      valid_str = attribute_str(current_node, "VALID");
      current_requirement->valid = (valid_str && xmlStrEqual(valid_str, "no")) ? false : true;

      /* subnet */
      if (current_node->parent->type == XML_ELEMENT_NODE
	  && xmlStrEqual(current_node->parent->name, tag_subnet)) {
	/* parent of current_node is a subnet node */
	xmlChar *subnet_id = attribute_str(current_node->parent, "ID");
	/* find the pointer to this subnet_t */
	subnet_t *s;
	for (s=pkg2->pn->sn; s->id; s++) {
	  if (xmlStrEqual(subnet_id, s->id)) {
	    /* same id */
	    current_requirement->sn = s;
	    /* break; */
	  }
	}
      }
      else {
	current_requirement->sn = pkg2->pn->sn;
      }

      /* inheritance of valid flag */
      if (! current_requirement->sn->valid) {
	current_requirement->valid = false;
      }

      /* search for the name of requirement -> children with content NAME */
      current_requirement->name =
	child_content(child_element(current_node, "NAME"));
      
      /* search for the description of state -> children with
	 content DESCRIPTION */
      current_requirement->description =
	child_content(child_element(current_node, "DESCRIPTION"));

      current_requirement->l = htlist_new(current_node->children, tag_hitem);

      current_requirement->check = htlist_new(current_node->children, tag_check);

      current_requirement++;
    }

    current_requirement =
      requirements_init(pkg2, current_requirement, current_node->children, element_name);
  }

  return current_requirement;
}
/* End of requirements_init() */


/*!

\param p pointer to the pkg2

\return success flag
*/
flag_t
requirement_refs_new(pkg2_t *pkg2)
{
  req_t *tail;

  assert(pkg2);
  assert(pkg2->pn->xml);
  assert(pkg2->pn->root_element);

  /* requirements */
  pkg2->n_req = number_of(pkg2->pn->root_element, tag_req, 0);

  printf_debug("Number of '%s' = %i", tag_req, pkg2->n_req);

  pkg2->req = (req_t *) mem_alloc(pkg2->n_req+1, sizeof(req_t));

  tail = reqs_init(pkg2, pkg2->req, pkg2->pn->root_element, tag_req);

  tail->ref = NULL;              /* mark the end of requirement array */

  return true;
}
/* End of pkg2_requirement_refs_new() */


/*!

\param p pointer to the pkg2

\return 
*/
req_t *
reqs_init(pkg2_t *pkg2, req_t *current_req, xmlNode *a_node, const xmlChar *element_name)
{
  xmlNode *current_node;
  xmlChar *valid_str;

  for (current_node = a_node; current_node; current_node = current_node->next) {

    if (current_node->type == XML_ELEMENT_NODE
	&& xmlStrEqual(current_node->name, element_name)) {
      /* current_node is a node w/ element_name */

      /* set valid flag */        
      valid_str = attribute_str(current_node, "VALID");
      current_req->valid = (valid_str && xmlStrEqual(valid_str, "no")) ? false : true;

      /* set ref of req */
      current_req->ref = attribute_str(current_node, "REF");

      /* set cat of req */
      current_req->cat = attribute_str(current_node, "CAT");

      /* find the parent of req in DOM */
      if (current_node->parent->type == XML_ELEMENT_NODE) {

	/* parent of current_node is a subnet node */
	xmlChar *id_str = attribute_str(current_node->parent, "ID");

	if (xmlStrEqual(current_node->parent->name, tag_state)) {
	  current_req->knot = state_ptr4id(pkg2->pn, id_str);
	  current_req->edge = NULL;
	}
	else if (xmlStrEqual(current_node->parent->name, tag_transition)) {
	  current_req->knot = transition_ptr4id(pkg2->pn, id_str);
	  current_req->edge = NULL;
	}
	else if (xmlStrEqual(current_node->parent->name, tag_edge)) {
	  /* find the pointer to this edge */
	  current_req->edge = edge_ptr4id(pkg2->pn, attribute_str(current_node->parent, "FROM"), attribute_str(current_node->parent, "TO"));
	  current_req->knot = NULL;
	}
      }


#if 0
      /* inheritance of valid flag */
      if (! current_req->sn->valid) {
	current_req->valid = false;
      }
#endif

      current_req++;
    }

    current_req =
      reqs_init(pkg2, current_req, current_node->children, element_name);
  }

  return current_req;
}
/* End of reqs_init() */


void
pkg2_free(pkg2_t *pkg2)
{
  index_t i;

  for (i=0; i < pkg2->n_r; i++) {
    if (pkg2->r[i].l) {
      mem_free(pkg2->r[i].l);
    }
    if (pkg2->r[i].check) {
      mem_free(pkg2->r[i].check);
    }
  }
  mem_free(pkg2->r);

  mem_free(pkg2->req);

  petrinet_free(pkg2->pn);

  mem_free(pkg2);

  return;
}
/* End of pkg2_free() */


/*! set all out flags in pkg2

*/
void
pkg2_all_out_flag(pkg2_t *pkg2, flag_t f)
{
  index_t r_i;

  petrinet_all_out_flag(pkg2->pn, f);

  /* out flag of requirements */
  for (r_i=0; r_i < pkg2->n_r; r_i++) {
    pkg2->r[r_i].out = f;
  }

  return;
}


/*! set all out flags for requirements in pkg2 according k
*/
void
pkg2_knot_requirement_out_flag(pkg2_t *pkg2, knot_t *k, flag_t f)
{
  index_t r_i;
  index_t e_i;
  index_t q_i;
  index_t k_i;
  knot_t *knots;

  for (r_i=0; r_i < pkg2->n_r; r_i++) {
    /* find the requirement w/ right id */
    if (pkg2->r[r_i].valid) {

      if (cat4knot(pkg2, k, &(pkg2->r[r_i])) != '-') {
	/*  */
	pkg2->r[r_i].out = f;
	continue;
      }

      for (k_i=0, knots = ((k->type == state) ? pkg2->pn->t : pkg2->pn->s);
	   knots[k_i].id;
	   k_i++) {
	if (knots[k_i].valid && knots[k_i].out
	    && (cat4edge(pkg2, k, &(knots[k_i]), &(pkg2->r[r_i])) != '-'
		|| cat4knot(pkg2, &(knots[k_i]), &(pkg2->r[r_i])) != '-')) {
	  pkg2->r[r_i].out = f;
	}
      }
    }
  }

  return;
}
/* Ende von pkg2_knot_requirement_out_flag() */


/*! cat4edge() - 
 */
char 
cat4edge(pkg2_t *pkg2, knot_t *k_a, knot_t *k_b, requirement_t *r)
{

  index_t q_i;
  index_t k_i;
  knot_t *knots;

  for (q_i=0; pkg2->req[q_i].ref; q_i++) { /* all reqs */
    if (pkg2->req[q_i].edge && xmlStrEqual(pkg2->req[q_i].ref, r->id)) {
      /* this requirement r[r_i] is refered by req[q_i] 
       */
      if (k_a->type == state) {
	if (pkg2->req[q_i].edge->from == k_a
	    && pkg2->req[q_i].edge->to == k_b) {
	  return pkg2->req[q_i].cat[0];
	}
      }
      else {
	if (pkg2->req[q_i].edge->from == k_b
	    && pkg2->req[q_i].edge->to == k_a) {
	  return pkg2->req[q_i].cat[0];
	}
      }
    }
  }

  return '-';
}
/* Ende von cat4edge() */


/*! cat4knot() - 
 */
char 
cat4knot(pkg2_t *pkg2, knot_t *k_a, requirement_t *r)
{

  index_t q_i;
  index_t k_i;
  knot_t *knots;

  for (q_i=0; pkg2->req[q_i].ref; q_i++) { /* all reqs */
    if (pkg2->req[q_i].knot 
	&& pkg2->req[q_i].knot == k_a
	&& xmlStrEqual(pkg2->req[q_i].ref, r->id)) {
      /* this requirement r[r_i] is refered by req[q_i] 
       */
      return pkg2->req[q_i].cat[0];
    }
  }

  return '-';
}
/* Ende von cat4knot() */


