/*

  Copyright (C) 1999-2005 by Alexander Tenbusch

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License 
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/


/*! handling functions

*/

#include <assert.h>
#include <stdlib.h>
#include <stdio.h>
#include <stddef.h>
#include <ctype.h>
#include <string.h>
#include <stdarg.h>

#include <libxml/parser.h>

#include "tools.h"
#include "petrinet.h"
#include "mem.h"


/*! converts an UTF8 string to iso latin 1

  \param b target pointer
  \param a source pointer
  \param n_max number of maximum characters

  \return target pointer
*/
unsigned char *
isolat1_str(unsigned char *b, xmlChar *a, size_t n_max)
{
  size_t l_a = xmlStrlen(a);

  assert(n_max > l_a);

  if (! UTF8Toisolat1( b, &n_max, a, &l_a) ) {
    n_max = 0;
  }

  b[n_max] = 0;

  return b;
}
/* End of isolat1_str() */


/*! counts the number of elements with right name

  \param a_node is the current node
  \param element_name string for name element
  \param n counter for recursive call

  \return the number of elements with name 'element_name'
*/
index_t
number_of(xmlNode * a_node, const xmlChar *element_name, index_t n)
{
  while (a_node) {
    if (a_node->type == XML_ELEMENT_NODE && xmlStrEqual(a_node->name, element_name))
      n++;

    n = number_of(a_node->children, element_name, n);

    a_node = a_node->next;
  }
  return n;
}


/*! searchs under the childs of 'a_node' for an element named 'element_name'

  \param a_node is the current node (for recursion)
  \param element_name string for name element

  \return the child node with name 'element_name'
*/
xmlNode *
child_element(xmlNode * a_node, const xmlChar *element_name)
{
  xmlNode *cur_node;

  for (cur_node = a_node->children; cur_node; cur_node = cur_node->next) {
    if (cur_node->type == XML_ELEMENT_NODE
	&& xmlStrEqual(cur_node->name, element_name)) {
      return cur_node;
    }
  }

  return NULL;
}
/* End of child_element() */


/*!

\return the text (the child content) of a node
*/
xmlChar *
child_content(xmlNode * a_node)
{
  if (a_node && a_node->children && a_node->children->type == XML_TEXT_NODE)
    return a_node->children->content;

  return empty_str;
}
/* End of child_content() */


/*!

\return the content pointer of the attribute named 'attribute_name'
of 'a_node' OR NULL if no attribute found
*/
xmlChar *
attribute_str(xmlNode * a_node, const xmlChar *attribute_name)
{
  xmlAttr *curr_attr;

  assert(a_node);
  assert(attribute_name);

  for (curr_attr = a_node->properties; curr_attr; curr_attr = curr_attr->next)
    if (curr_attr->type == XML_ATTRIBUTE_NODE
	&& xmlStrEqual(curr_attr->name, attribute_name))
      return curr_attr->children->content;
    
  return NULL;
}
/* End of attribute_str() */


/*!

\param a_node current node

\return pointer to created htitem array
*/
htitem_t *
htlist_new(xmlNode * a_node, const xmlChar *tag)
{
  htitem_t *result = NULL;

  index_t i;
  index_t n;

  assert(a_node);

  /* htitems */
  n = number_of(a_node, tag, 0);

  if (n > 0) {

/*     printf_debug("Number of '%s' = %i", tag, n); */

    result = (htitem_t *) mem_alloc(n+1, sizeof(htitem_t));

    for (i=0; a_node; ) {

      if (a_node->type == XML_ELEMENT_NODE && xmlStrEqual(a_node->name, tag)) {
	/* a_node has the right tag */

	assert(i < n);

	/* set ref of htitem */
	result[i].href = attribute_str(a_node, "HREF");

	/* set text of htitem */
	result[i].text = child_content(a_node);

	i++;
      }
      a_node = a_node->next;
    }
    result[i].text = NULL; 	/* marker for last element */
  }

  return result;
}
/* End of htlist_new() */


void
printf_exit(flag_t flag, int exit_code, char *fmt, ...)
{
  /* 
     fprintf's error messages and exit(exit_code)
  */

  if (flag) {
    va_list ap;

    va_start(ap,fmt);

    fprintf(stderr,"! ERROR '%i': ", exit_code);
    vfprintf(stderr, fmt, ap);
    fprintf(stderr," ...\n");

    va_end(ap);

    exit(exit_code);
  }
}


void
printf_debug(char *fmt, ...)
{
  /* 
     fprintf's debug messages (in DEBUG mode only!)
  */

#if defined DEBUG
  va_list ap;

  va_start(ap,fmt);

  fprintf(stderr,"! DEBUG: ");
  vfprintf(stderr, fmt, ap);
  fprintf(stderr," ...\n");

/*   fprintf(stdout,"<pre>! DEBUG: "); */
/*   vfprintf(stdout, fmt, ap); */
/*   fprintf(stdout," ...</pre>"); */

  va_end(ap);

#endif
}

/*! concates a variable number of strings into a buffer
 */
char *
sprintpath(char *buffer, ...)
{
  va_list ap;
  char *e;

  va_start(ap,buffer);

  while ((e = va_arg(ap,char*))) {
    fputs(e,stdout);
  }

  va_end(ap);

  return buffer;
}

